from ._FromTensorflowExample import FromTensorflowExample
from ._TensorflowExampleReader import TensorflowExampleReader
from ._TensorflowExampleWriter import TensorflowExampleWriter
from ._ToTensorflowExample import ToTensorflowExample



import io
from typing import Tuple

import numpy as np
import PIL
from wai.common.geometry import Polygon, Point

from wai.annotations.image_utils import mask_to_polygon


def polygon_from_mask(mask: bytes,
                      view: Tuple[float, float, float, float],
                      mask_threshold: float = 0.9,
                      sample_stride: int = 1) -> Polygon:
    """
    Converts the PNG format mask into a polygon.

    :param mask:            The PNG mask.
    :param view:            The area inside the mask to look for the polygon (left, top, right, bottom).
    :param mask_threshold:  The threshold to consider the mask inside the polygon.
    :param sample_stride:   The stride to step through the mask when finding the polygon. Results
                            in a more coarsely-grained polygon but finishes faster.
    :return:                A polygon.
    """
    # Validate mask_threshold parameter
    if not (0.0 <= mask_threshold <= 1.0):
        raise ValueError(f"Mask threshold must be in [0.0, 1.0] but got {mask_threshold}")

    # Decode the PNG
    image = PIL.Image.open(io.BytesIO(mask))

    # Convert to a bitmask
    binary_mask = np.asarray(image)

    # Use find_contours to find the (only) polygon in the bit-mask
    poly_array = mask_to_polygon(binary_mask,
                                 mask_threshold=mask_threshold,
                                 view=view,
                                 mask_nth=sample_stride,
                                 fully_connected="high")[0]

    # Convert to a Polygon
    polygon = Polygon()
    for p in range(poly_array.shape[0]):
        polygon.points.append(Point(int(round(poly_array[p, 1])), int(round(poly_array[p, 0]))))

    # Remove redundant line segments
    polygon.simplify()

    return polygon



from typing import List, Union

from ._ensure_available import tensorflow as tf


def extract_feature(features: tf.train.Features, name: str) -> Union[List[bytes], List[int], List[float]]:
    """
    Extracts the feature with the given name from the given features.

    :param features:    The tensorflow features to extract from.
    :param name:        The name of the feature to extract.
    :return:            The feature value list, or an empty list
                        if the feature isn't present in the features.
    """
    # Check the feature exists
    if name not in features.feature:
        return []

    # Select the feature
    feature = features.feature[name]

    if len(feature.bytes_list.value) > 0:
        return list(map(bytes, feature.bytes_list.value))
    elif len(feature.int64_list.value) > 0:
        return list(map(int, feature.int64_list.value))
    elif len(feature.float_list.value) > 0:
        return list(map(float, feature.float_list.value))
    else:
        return []



import numpy as np


def dense_numerical_from_mask(mask: np.ndarray) -> np.ndarray:
    """
    Encodes a binary image mask in Tensorflow's dense numerical format, in place.

    :param mask:    The image mask.
    :return:        The encoded array.
    """
    mask.resize((mask.size,))
    return mask



from typing import Type, Tuple

from wai.annotations.core.component import Component
from wai.annotations.core.domain import DomainSpecifier
from wai.annotations.core.specifier import SourceStageSpecifier


class TFRecordsInputFormatSpecifier(SourceStageSpecifier):
    """
    Specifier of the components for reading the binary TFRecords
    object detection format.
    """
    @classmethod
    def description(cls) -> str:
        return "Reads image object-detection annotations in the TFRecords binary format"

    @classmethod
    def components(cls) -> Tuple[Type[Component], ...]:
        from wai.annotations.core.component.util import LocalFilenameSource
        from ..component import TensorflowExampleReader, FromTensorflowExample
        return LocalFilenameSource, TensorflowExampleReader, FromTensorflowExample

    @classmethod
    def domain(cls) -> Type[DomainSpecifier]:
        from wai.annotations.domain.image.object_detection import ImageObjectDetectionDomainSpecifier
        return ImageObjectDetectionDomainSpecifier



"""
Compatibility layer for working with Tensorflow v1 and v2.
"""
# Get the version of Tensorflow
try:
    import tensorflow
    VERSION = tensorflow.version.VERSION
except Exception as e:
    raise RuntimeError(f"Failed to determine Tensorflow version") from e

# Cast the major version number to an int
try:
    version = int(VERSION[0])
except Exception as e:
    raise RuntimeError(f"Could not parse Tensorflow major version number from '{VERSION}'") from e

# Import the correct compatibility package
if version == 1:
    from .v1 import *
elif version == 2:
    from .v2 import *
else:
    raise RuntimeError(f"Major version number parsed from Tensorflow version ('{VERSION}') "
                       f"was {version}, but should be 1 or 2")



from typing import List

from wai.common.adams.imaging.locateobjects import LocatedObjects, LocatedObject
from wai.common.cli.options import TypedOption

from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.stream import ThenFunction, DoneFunction
from wai.annotations.core.stream.util import RequiresNoFinalisation
from wai.annotations.domain.image import Image
from wai.annotations.domain.image.object_detection import ImageObjectDetectionInstance
from wai.annotations.domain.image.object_detection.util import set_object_label
from ..utils import (
    extract_feature,
    image_info_from_example,
    polygon_from_mask,
    TensorflowExampleExternalFormat
)


class FromTensorflowExample(
    RequiresNoFinalisation,
    ProcessorComponent[TensorflowExampleExternalFormat, ImageObjectDetectionInstance]
):
    """
    Converter from Tensorflow Examples to the internal format.
    """
    mask_threshold = TypedOption(
        "--mask-threshold",
        type=float,
        default=0.9,
        metavar="THRESHOLD",
        help="the threshold to use when calculating polygons from masks"
    )

    sample_stride = TypedOption(
        "--sample-stride",
        type=int,
        default=1,
        metavar="STRIDE",
        help="the stride to use when calculating polygons from masks"
    )

    def process_element(
            self,
            element: TensorflowExampleExternalFormat,
            then: ThenFunction[ImageObjectDetectionInstance],
            done: DoneFunction
    ):
        # Define a UTF-8 decoder
        def decode_utf_8(bytes_: bytes) -> str:
            return bytes_.decode("utf-8")

        # Extract the image data from the example instance
        image_info: Image = image_info_from_example(element)

        # Extract the located object data from the instance
        lefts = extract_feature(element.features, 'image/object/bbox/xmin')
        rights = extract_feature(element.features, 'image/object/bbox/xmax')
        tops = extract_feature(element.features, 'image/object/bbox/ymin')
        bottoms = extract_feature(element.features, 'image/object/bbox/ymax')
        labels = list(map(decode_utf_8, extract_feature(element.features, 'image/object/class/text')))

        # Extract the optional mask information
        masks = extract_feature(element.features, 'image/object/mask')
        if len(masks) == 0:
            masks = [b""] * len(lefts)

        located_objects = self.process_located_objects(lefts, rights, tops, bottoms,
                                                       labels, masks,
                                                       image_info.width, image_info.height)

        then(ImageObjectDetectionInstance(image_info, located_objects))

    def process_located_objects(self,
                                lefts: List[float],
                                rights: List[float],
                                tops: List[float],
                                bottoms: List[float],
                                labels: List[str],
                                masks: List[bytes],
                                image_width: int,
                                image_height: int) -> LocatedObjects:
        """
        Processes the data from the Tensorflow example into a list of located objects.

        :param lefts:           The left bounds of the objects.
        :param rights:          The right bounds of the objects.
        :param tops:            The top bounds of the objects.
        :param bottoms:         The bottom bounds of the objects.
        :param labels:          The labels of the objects.
        :param masks:           The masks of the objects.
        :param image_width:     The width of the image.
        :param image_height:    The height of the image.
        :return:                The located objects.
        """
        # Create the empty located objects
        located_objects: LocatedObjects = LocatedObjects()

        # Process each object in turn
        for left, right, top, bottom, label, mask in zip(lefts, rights, tops, bottoms, labels, masks):
            # Calculate the denormalised coordinates of the object
            x = round(left * image_width)
            y = round(top * image_height)
            width = round(right * image_width) - x + 1
            height = round(bottom * image_height) - y + 1

            # Create the located object
            located_object: LocatedObject = LocatedObject(x, y, width, height)
            set_object_label(located_object, label)

            if len(mask) != 0:
                located_object.set_polygon(
                    polygon_from_mask(
                        mask,
                        (left * image_width, top * image_height, right * image_width, bottom * image_height),
                        self.mask_threshold,
                        self.sample_stride
                    )
                )

            # Add the object to the list
            located_objects.append(located_object)

        return located_objects



import hashlib
from typing import Tuple, List, Dict

import numpy as np

from wai.common.adams.imaging.locateobjects import LocatedObjects
from wai.common.cli.options import FlagOption, TypedOption

from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.stream import ThenFunction, DoneFunction
from wai.annotations.core.stream.util import ProcessState, RequiresNoFinalisation
from wai.annotations.domain.image.object_detection import ImageObjectDetectionInstance
from wai.annotations.domain.image.object_detection.util import get_object_label
from wai.annotations.domain.image.segmentation.util import mask_from_polygon
from ..utils import (
    tensorflow as tf,
    make_feature,
    negative_example,
    TensorflowExampleExternalFormat,
    png_from_mask,
    dense_numerical_from_mask
)


class ToTensorflowExample(
    RequiresNoFinalisation,
    ProcessorComponent[ImageObjectDetectionInstance, TensorflowExampleExternalFormat]
):
    """
    Converter from the internal format to Tensorflow Examples.
    """
    dense_format: bool = FlagOption(
        "--dense",
        help="outputs masks in the dense numerical format instead of PNG-encoded"
    )

    source_id_type: bool = TypedOption(
        "--source-id-type",
        type=str,
        choices=["filename", "numeric-dummy"],
        help="by default, the filename gets stored in the 'source_id' field, but some algorithms try to convert it "
             + "into a number and fail with 'StringToNumberOp could not correctly convert string'; in which case you "
             + "can use 'numeric-dummy' (see https://github.com/google/automl/issues/307)",
        default="filename"
    )

    _label_class_lookup: Dict[str, int] = ProcessState(lambda self: {})

    def process_element(
            self,
            element: ImageObjectDetectionInstance,
            then: ThenFunction[TensorflowExampleExternalFormat],
            done: DoneFunction
    ):
        image_info, located_objects = element

        # Make sure we have an image
        if image_info.data is None:
            raise ValueError(f"Tensorflow records require image data")

        # If no annotations, return an empty example
        if located_objects is None or len(located_objects) == 0:
            return then(negative_example(image_info))

        # Format and extract the relevant annotation parameters
        lefts, rights, tops, bottoms, labels, classes, masks, is_crowds, areas = \
            self.process_located_objects(located_objects, image_info.width, image_info.height)

        # generate the source_id
        if self.source_id_type == "filename":
            source_id = image_info.filename.encode("utf-8")
        elif self.source_id_type == "numeric-dummy":
            source_id = "0".encode("utf-8")
        else:
            raise Exception("Unhandled source_id type: %s" % self.source_id_type)

        # Create the example features
        feature_dict = {
            'image/height': make_feature(image_info.height),
            'image/width': make_feature(image_info.width),
            'image/filename': make_feature(image_info.filename.encode("utf-8")),
            'image/source_id': source_id,
            'image/encoded': make_feature(image_info.data),
            'image/format': make_feature(image_info.format.get_default_extension().encode("utf-8")),
            'image/key/sha256': make_feature(hashlib.sha256(image_info.data).hexdigest().encode("utf-8")),
            'image/object/bbox/xmin': make_feature(lefts),
            'image/object/bbox/xmax': make_feature(rights),
            'image/object/bbox/ymin': make_feature(tops),
            'image/object/bbox/ymax': make_feature(bottoms),
            'image/object/class/text': make_feature(labels),
            'image/object/class/label': make_feature(classes),
            'image/object/is_crowd': make_feature([1 if is_crowd else 0 for is_crowd in is_crowds]),
            'image/object/area': make_feature(areas)
        }

        # Add the masks if present
        if len(masks) > 0:
            # Encode the masks based on the --dense option
            feature_dict['image/object/mask'] = (
                tf.train.Feature(float_list=tf.train.FloatList(value=np.concatenate(list(map(dense_numerical_from_mask, masks)))))
                if self.dense_format else
                make_feature(list(map(png_from_mask, masks)))
            )

        # Create and return the example
        then(
            tf.train.Example(
                features=tf.train.Features(
                    feature=feature_dict
                )
            )
        )

    def process_located_objects(self, located_objects: LocatedObjects, image_width: int, image_height: int) -> Tuple[
        List[float],
        List[float],
        List[float],
        List[float],
        List[bytes],
        List[int],
        List[np.ndarray],
        List[bool],
        List[float]
    ]:
        """
        Processes the located objects into the format expected by Features.

        :param located_objects:     The located objects.
        :param image_width:         The width of the image.
        :param image_height:        The height of the image.
        :return:                    A tuple of lists of:
                                        - left bounds
                                        - right bounds
                                        - top bounds
                                        - bottom bounds
                                        - UTF-8 encoded class labels
                                        - class categories
                                        - masks
                                        - is_crowd flags (always false)
                                        - areas
        """
        # Format and extract the relevant annotation parameters
        lefts = []
        rights = []
        tops = []
        bottoms = []
        labels = []
        classes = []
        masks = []
        is_crowds = []
        areas = []
        for located_object in located_objects:
            # Get the object label
            label = get_object_label(located_object)

            # Skip unknown labels if given a specific set, or add it
            # if using auto-labeling
            if label not in self._label_class_lookup:
                self._label_class_lookup[label] = len(self._label_class_lookup) + 1

            # Get the class
            class_ = self._label_class_lookup[label]

            # Normalise the boundary coordinates
            left = located_object.x / image_width
            right = (located_object.x + located_object.width - 1) / image_width
            top = located_object.y / image_height
            bottom = (located_object.y + located_object.height - 1) / image_height

            # Append the object to the lists if its kosher
            if (0.0 <= left < right <= 1.0) and (0.0 <= top < bottom <= 1.0):
                lefts.append(left)
                rights.append(right)
                tops.append(top)
                bottoms.append(bottom)
                labels.append(label.encode('utf-8'))
                classes.append(class_)
                is_crowds.append(False)
                if located_object.has_polygon():
                    polygon = located_object.get_polygon()
                    masks.append(
                        mask_from_polygon(
                            polygon,
                            image_width,
                            image_height
                        )
                    )
                    areas.append(polygon.area())
                else:
                    # TODO: If some located objects have polygons and some don't,
                    #       the 'masks' list will be a different size to the others.
                    #       Resolve.
                    areas.append(float(located_object.get_rectangle().area()))

        return lefts, rights, tops, bottoms, labels, classes, masks, is_crowds, areas



from ..utils import tensorflow as tf

from wai.annotations.core.component.util import AnnotationFileProcessor
from wai.annotations.core.stream import ThenFunction
from wai.annotations.domain.image import Image
from ..utils import negative_example, TensorflowExampleExternalFormat


class TensorflowExampleReader(AnnotationFileProcessor[TensorflowExampleExternalFormat]):
    """
    Reader of Tensorflow Example records.
    """
    def read_annotation_file(self, filename: str, then: ThenFunction[TensorflowExampleExternalFormat]):
        # Load the record file
        dataset = tf.data.TFRecordDataset([filename])

        # Extract each example and yield
        for record in dataset:
            # Create the instance
            instance: TensorflowExampleExternalFormat = tf.train.Example()

            # Parse the record
            instance.ParseFromString(record.numpy())

            then(instance)

    def read_negative_file(self, filename: str, then: ThenFunction[TensorflowExampleExternalFormat]):
        then(negative_example(Image.from_file(filename)))



def ensure_eager_execution():
    """
    Ensures that Tensorflow is in eager-execution mode.
    """
    # Tensorflow V2 is eager by default
    from tensorflow import executing_eagerly
    if not executing_eagerly():
        raise RuntimeError(f"Tensorflow V2 not in eager-execution mode")



from typing import List, Dict

from ..utils import TensorflowExampleExternalFormat
from ._extract_feature import extract_feature


class LabelMapAccumulator:
    """
    Accumulates a label map from the passing Tensorflow examples.
    """
    # Property to get the label map
    label_map = property(lambda self: self._label_map.copy())

    def __init__(self):
        # The accumulated label map
        self._label_map: Dict[str, int] = {}

    def accumulate(self, example: TensorflowExampleExternalFormat):
        # Create a method to decode the binary strings
        def decode_utf_8(bytes_: bytes) -> str:
            return bytes_.decode("utf-8")

        # Get and decode the labels from the example
        labels: List[str] = list(map(decode_utf_8, extract_feature(example.features, 'image/object/class/text')))

        # Get the classes from the example
        classes: List[int] = extract_feature(example.features, 'image/object/class/label')

        # Make sure the labels and classes are the same size
        if len(labels) != len(classes):
            raise ValueError(f"Number of labels differs from number of classes")

        # Check and add labels/classes to the label map
        for label, class_ in zip(labels, classes):
            if label in self._label_map:
                if self._label_map[label] != class_:
                    raise RuntimeError(f"'{label}' maps to {self._label_map[label]} already but was "
                                       f"redefined to {class_}")
            else:
                self._label_map[label] = class_



from wai.annotations.domain.image import Image, ImageFormat
from ..utils import TensorflowExampleExternalFormat
from ._extract_feature import extract_feature


def image_info_from_example(instance: TensorflowExampleExternalFormat):
    """
    Extracts an image-info object from a Tensorflow example.

    :param instance:    The Tensorflow example.
    :return:            The image-info.
    """
    image_filename = extract_feature(instance.features, 'image/filename')[0].decode("utf-8")
    image_data = extract_feature(instance.features, 'image/encoded')[0]
    image_width = extract_feature(instance.features, 'image/width')[0]
    image_height = extract_feature(instance.features, 'image/height')[0]
    image_format = ImageFormat.for_extension(extract_feature(instance.features, 'image/format')[0].decode("utf-8"))

    if not isinstance(image_width, int) or not isinstance(image_height, int):
        raise TypeError(f"Image dimensions not in expected format (integer)")

    return Image(image_filename, image_data, image_format, (image_width, image_height))



"""
Package for utilities for working with the Tensorflow Records format.
"""
from ._dense_numerical_from_mask import dense_numerical_from_mask
from ._extract_feature import extract_feature
from ._ensure_available import tensorflow
from ._format import TensorflowExampleExternalFormat
from ._image_info_from_example import image_info_from_example
from ._LabelMapAccumulator import LabelMapAccumulator
from ._make_feature import make_feature
from ._negative_example import negative_example
from ._png_from_mask import png_from_mask
from ._polygon_from_mask import polygon_from_mask



from typing import List, Union

from ._ensure_available import tensorflow as tf


def make_feature(value: Union[bytes, int, float, List[bytes], List[int], List[float]]) \
        -> tf.train.Feature:
    """
    Creates a feature of the correct format for the given value.

    :param value:   The value to create a feature for. Can be a
                    string, bytes, int or float, or a list of such.
    :return:        The feature.
    """
    # Convert scalars into lists
    if not isinstance(value, list):
        value = [value]

    # Must have at least one element in the list
    if len(value) == 0:
        raise Exception("make_feature cannot handle length-zero lists, feature must be created manually")

    # Determine the argument and class to Feature from the element type
    if isinstance(value[0], int):
        arg_name, cls = "int64_list", tf.train.Int64List
    elif isinstance(value[0], bytes):
        arg_name, cls = "bytes_list", tf.train.BytesList
    elif isinstance(value[0], float):
        arg_name, cls = "float_list", tf.train.FloatList
    else:
        raise Exception(f"make_feature only handles int/bytes/float types, got {type(value[0])}")

    return tf.train.Feature(**{arg_name: cls(value=value)})



import hashlib

from wai.annotations.domain.image import Image
from ._ensure_available import tensorflow as tf
from ._make_feature import make_feature


def negative_example(image_info: Image):
    """
    Creates an empty example (for images with no annotations).

    :return:    The empty example.
    """
    return tf.train.Example(
        features=tf.train.Features(
            feature={
                'image/height': make_feature(image_info.height),
                'image/width': make_feature(image_info.width),
                'image/filename': make_feature(image_info.filename.encode("utf-8")),
                'image/source_id': make_feature(image_info.filename.encode("utf-8")),
                'image/encoded': make_feature(image_info.data),
                'image/format': make_feature(image_info.format.get_default_extension().encode("utf-8")),
                'image/key/sha256': make_feature(hashlib.sha256(image_info.data).hexdigest().encode("utf-8")),
                'image/object/bbox/xmin': tf.train.Feature(float_list=tf.train.FloatList(value=[])),
                'image/object/bbox/xmax': tf.train.Feature(float_list=tf.train.FloatList(value=[])),
                'image/object/bbox/ymin': tf.train.Feature(float_list=tf.train.FloatList(value=[])),
                'image/object/bbox/ymax': tf.train.Feature(float_list=tf.train.FloatList(value=[])),
                'image/object/class/text': tf.train.Feature(bytes_list=tf.train.BytesList(value=[])),
                'image/object/class/label': tf.train.Feature(int64_list=tf.train.Int64List(value=[])),
                'image/object/is_crowd': tf.train.Feature(int64_list=tf.train.Int64List(value=[])),
                'image/object/area': tf.train.Feature(float_list=tf.train.FloatList(value=[])),
            }
        )
    )



import os
from os.path import join
from typing import List, Optional, Tuple, Iterator, ContextManager

from wai.common.cli.options import TypedOption

from wai.annotations.core.component.util import (
    SplitSink,
    LocalFileWriter,
    WithPersistentSplitFiles,
    ExpectsFile
)
from wai.annotations.core.stream.util import ProcessState
from wai.annotations.core.util import InstanceState
from ..utils import (
    tensorflow as tf,
    LabelMapAccumulator,
    TensorflowExampleExternalFormat
)


class TensorflowExampleWriter(
    ExpectsFile,
    WithPersistentSplitFiles[List[tf.io.TFRecordWriter]],
    LocalFileWriter[Tuple[int, TensorflowExampleExternalFormat]],
    SplitSink[Tuple[int, TensorflowExampleExternalFormat]]
):
    """
    Writer of Tensorflow example records.
    """
    shards = TypedOption(
        "-s", "--shards",
        type=str,
        nargs="+",
        metavar="FILENAME",
        help="additional shards to write to"
    )

    protobuf_label_map = TypedOption(
        "-p", "--protobuf",
        type=str,
        metavar="FILENAME",
        help="for storing the label strings and IDs"
    )

    _num_writers: int = InstanceState(lambda self: len(self.shards) + 1)

    _label_map_accumulator: Optional[LabelMapAccumulator] = ProcessState(
        lambda self: None if self.protobuf_label_map is None else LabelMapAccumulator()
    )

    def consume_element_for_split(
            self,
            element: Tuple[int, TensorflowExampleExternalFormat]
    ):
        index, example = element

        if self._label_map_accumulator is not None:
            self._label_map_accumulator.accumulate(example)

        self._split_files[index % self._num_writers].write(example.SerializeToString())

    def finish_split(self):
        if self._label_map_accumulator is None:
            return

        # Write the accumulated labels
        self.write_protobuf_label_map(
            self._label_map_accumulator,
            self.get_split_path(self.split_label, self.output_path)
        )

    def write_protobuf_label_map(self, label_map_accumulator: LabelMapAccumulator, path: str):
        """
        Writes the label-to-index mapping to the given file, in
        protobuf format, if an output filename was given.
        """
        # Extract the label map and sort it
        labels_sorted = [
            (index, label)
            for label, index in label_map_accumulator.label_map.items()
        ]
        labels_sorted.sort(key=lambda entry: entry[0])

        # Format the label index map
        protobuf: List[str] = [
            "item {\n" +
            f"  id: {index}\n" +
            f"  name: '{label}'\n" +
            "}\n"
            for index, label in labels_sorted
        ]

        # Write the lines to the specified file
        with open(join(path, self.protobuf_label_map), 'w') as file:
            file.writelines(protobuf)

    @classmethod
    def get_help_text_for_output_option(cls) -> str:
        return "name of output file for TFRecords"

    def _init_split_files(self) -> List[tf.io.TFRecordWriter]:
        # Get the split path
        split_path = self.get_split_path(self.split_label, self.output, True)

        # Format the paths for the shards
        all_file_names = [split_path]
        for shard in self.shards:
            # Concatenate all output file-names
            all_file_names.append(
                os.path.join(
                    os.path.dirname(split_path),
                    shard
                )
            )

        # Open the output file for sharded writing
        return [tf.io.TFRecordWriter(file_name) for file_name in all_file_names]

    def _iterate_split_files(self, split_files: List[tf.io.TFRecordWriter]) -> Iterator[ContextManager]:
        return iter(split_files)



from io import BytesIO

import numpy as np

import PIL


def png_from_mask(mask: np.ndarray) -> bytes:
    """
    Encodes a binary image mask as a PNG file.

    :param mask:    The image mask to encode.
    :return:        The PNG file data.
    """
    # Write the bitmask into PNG format
    output_io = BytesIO()
    image = PIL.Image.fromarray(mask)
    image.save(output_io, format="PNG")

    return output_io.getvalue()



"""
Makes sure tensorflow is available. To utilise, just import this module.
"""
try:
    import tensorflow
except ImportError as e:
    raise RuntimeError("No tensorflow library found\n"
                       "Please install either tensorflow or tensorflow-gpu\n"
                       "    pip install tensorflow\n"
                       "    pip install tensorflow-gpu") from e

# Ensure we're in eager execution mode
from ..compat import ensure_eager_execution
ensure_eager_execution()
